(* Delphi-Unit
   verschiedene Unterroutinen
   ==========================
   - String-Konvertierungen ANSI <-> OEM
   - div. String-Routinen (z.B. Leerezeichen anfgen oder lschen)
   - Dateipfade zerlegen und zusammensetzen
   - Befehlszeilen auswerten
   - Umwandlung Grad:Min:Sek -> String und zurck
   - Substrings und Zahlen bis Begrenzer extrahieren
   - Gltigkeitsprfung fr Zahlen
   - Umwandlung Integer und Double -> String mit fhrenden Nullen und Tausendermark.
   - Dateigre in String
   - Zusammensetzen von Word und Cardinal aus Bytes
   - Typdeklarationen fr das Zerlegen von Word, LongWort und Int64 in Bytes und Words
   - schnelles Sortieren einer Liste (Quicksort)

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Neue Zusammenstellung - Mai. 2007
   letzte nderung       - Mrz. 2008
   *)

unit JrUtils;

interface

uses
  Windows, SysUtils, Classes, Graphics, Forms, Controls, ExtCtrls,
  Dialogs, Messages, IniFiles, Printers, ComCtrls;

type
  DateStr = string[10];
  TCvTable = array [0..127] of char;

  // Compare function for Quicksort
  TCompFunction = function (const arg1,arg2) : boolean of object;

  TDelimiters = set of char;

const
  ChrID = '#';             // Kennzeichner fr num. Zeichenwerte
  Strich = '-';
  Komma = ',';
  Punkt = '.';
  BSlash = '\';
  DPunkt = ':';
  Semikolon = ';';
  DecPt = '.';
  VertBar = '|';
  Quote = '"';
  Space = ' ';
  Tab = #9;
  CrLf = #13#10;
  MaxInt64 = High(int64);

  DezTr : char = Punkt;    // Dezimalzeichen
  ThMark : char = Space;   // Zeichen fr Tausendergruppen
  TmMark : char = DPunkt;  // Zeichen fr Uhrzeitgruppen hh:mm:ss

  xDef = 1.0;              // Standardwert ReadNxtDbl
  nDef = 1;                // Standardwert ReadNxtInt
  cDef = #0;               // Standardwert ReadNxtChr

  TabToANSI : TCvTable =(
    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,#$EA,#$EB,#$E9,#$E8,#$EF,#$EE,#$C4,#$C5,
    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,#$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$20,#$20,
    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$61,#$6F,#$BF,#$20,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
    #$20,#$20,#$20,#$A6,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$AD,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$DF,#$B6,#$20,#$20,#$20,#$B5,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$B1,#$20,#$20,#$20,#$20,#$F7,#$20,#$B0,#$20,#$B7,#$20,#$B3,#$B2,#$20,#$20);
  TabToOEM : TCvTable =(
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$AD,#$9B,#$9C,#$20,#$9D,#$B3,#$15,#$20,#$20,#$20,#$AE,#$AA,#$C4,#$20,#$20,
    #$F8,#$F1,#$FD,#$FC,#$20,#$E6,#$E3,#$20,#$20,#$20,#$20,#$AF,#$AC,#$AB,#$20,#$A8,
    #$41,#$41,#$41,#$41,#$8E,#$8F,#$92,#$80,#$45,#$90,#$45,#$45,#$49,#$49,#$49,#$49,
    #$44,#$A5,#$4F,#$4F,#$4F,#$4F,#$99,#$20,#$20,#$55,#$55,#$55,#$9A,#$59,#$20,#$E1,
    #$85,#$61,#$83,#$61,#$84,#$86,#$91,#$87,#$8A,#$82,#$88,#$89,#$8D,#$A1,#$8C,#$8B,
    #$20,#$A4,#$95,#$A2,#$93,#$6F,#$94,#$F6,#$20,#$97,#$A3,#$96,#$81,#$79,#$20,#$98);

  IllegalFilenameChars = [#0..#31,'*','?','<','>',':','"','/','\','|'];

  CenterPos : TPoint = (X : -1; Y : -1);
  MaxHist : integer = 50;

type
  TBoolFunction = function : boolean of object;
  TIntegerFunction = function : integer;

  TWinkelFormat = (fmDezGrad,fmDezMin,fmSekunden);
  TTextChange = (tcNone,tcLower,tcUpper);

  TWord = record
    case integer of
    0 : (Both : word);
    1 : (Lo,Hi : byte);
    end;

  TLongWord = record
    case integer of
    0 : (LongWord : cardinal);
    1 : (Lo,Hi : word);
    2 : (LoL,LoH,HiL,HiH : byte);
    3 : (Bytes : array [0..3] of Byte);
    end;

  TInt64 = record
    case integer of
    0: (AsInt64 : int64);
    1: (Lo, Hi  : Cardinal);
    2: (Cardinals: array [0..1] of Cardinal);
    3: (Words: array [0..3] of Word);
    4: (Bytes: array [0..7] of Byte);
    end;

  TFontStyleToByte = record
    case integer of
    1 : (Style : TFontStyles);
    2 : (Value : byte);
    end;

  TFPoint = record
    X,Y : double;
    end;

  TFRect = record
    case integer of
    0 : (Left,Top,Right,Bottom : double);
    1 : (TopLeft,BottomRight : TFPoint);
    end;

{ ---------------------------------------------------------------- }
(* Verzgerungszeit unter Verwendung der Windows Waitable Timer Objects
   Wait: warten bis Zeit in ms abgelaufen ist (min. Zeit 100 ns) *)

  TWaitTimer = class (TObject)
    FTimer   : THandle;
    constructor Create;
    destructor Destroy;  override;
    procedure Wait (DelayTime : double);
    end;

{ ---------------------------------------------------------------- }
// Anzeige eines Hinweisefenster (THintWindow), das nach einstellbarer Zeit (Delay)
// automatisch verschwindet
  TTimerHint = class (THintWindow)
  private
    FTimer : TTimer;
    FOnTerminate : TNotifyEvent;
    procedure Terminate (Sender : TObject);
  public
    constructor Create (AOwner: TComponent; Delay : integer);  // Delay in ms
    destructor Destroy; override;
    procedure ShowHint (r : TRect; AHint : string);
    procedure HideHint;
    property OnTerminate : TNotifyEvent read FOnTerminate write FOnTerminate;
    end;


{ ---------------------------------------------------------------- }
// Umwandeln eines DOS-Zeichens in eines Grobuchstaben (auch Umlaute)
function OEMUpCase (c : char) : char;

// Umwandeln eines ANSI-Zeichens in eines Grobuchstaben (auch Umlaute)
function ANSIUpCase (c : char) : char;

// Umwandeln eines DOS-Strings in Grobuchstaben (auch Umlaute) 
function OEMUpString (const S : string) : string;

// Umwandeln eines ANSI-Strings in Grobuchstaben (auch Umlaute) 
function ANSIUpString (const S : string) : string;

// Umwandeln eines Strings in Grobuchstaben (Umlaute in AE, OE, UE) zu Sortierzwecken
function OEMUpSort (const S : string) : string;

// Umwandeln eines Zeichens in Kleinbuchstaben
function LowCase(Ch: Char): Char;

// Umwandeln eines OEM-Strings nach ANSI
function StrToAnsi (const s : string) : string;

// Umwandeln eines ANSI-Strings nach OEM 
function StrToOEM (const s : string) : string;

// Substring-Pos. ohne Unterschreidung gro/klein
function TextPos (const Substr, S: string): Integer;
function TextPosEx (const Substr, S: string; Offset : cardinal): Integer;

// Text auf Gro- oder Kleinschreibung ndern
function TextChangeCase(s: string; CaseMode : TTextChange) : string;

// Info-String aus zwei Zeilen erzeugen
function MakeInfoStr (s1,s2 : string) : string;

{ ---------------------------------------------------------------- }
// String mit Leerstellen erzeugen
function FillSpace (len : integer) : string;

// fhrende Leerstellen ergnzen
function AddSp (const S : string;
                len     : integer) : string;

// Leerstellen am Ende ergnzen
function ExtSp (const S : string;
                len   : integer) : string;

// fhrende Leerstellen lschen
function RemSp (const S : string) : string;

// Leerstellen am Ende lschen
function CutSp (const S : string) : string;

// alle Leerstellen im String lschen
function DelSp (const S : string) : string;

// mehrfache Leerstellen im String lschen
function DelMulSp (const S : string) : string;

// Steuerzeichen im String lschen bzw. ersetzen
function DelCtrlChars(const S : string) : string;
function CrLfToChar (s : string; c : char) : string;
function CharToCr (s : string; c : char) : string;

// erstes nichtleeres Zeichen suchen
function SeekNoSp (const s : string) : integer;

// nchstes Leerzeichen oder Tab suchen
function SeekSpace (const s : string) : integer;

// String mit Zeichen erzeugen
function FillStr (c   : char;
                  len : integer) : string;

// fhrende Zeichen ergnzen
function AddChar (const S : string;
                  c       : char;
                  len     : integer) : string;

// Zeichen am Ende ergnzen
function ExtChar (const S : string;
                  c       : char;
                  len     : integer) : string;

// ein Zeichen durch ein anderes ersetzen
function ReplChars (const s         : string;
                   OldChar,NewChar : char) : string;

// fhrende Zeichen lschen
function RemChar (const S : string; c : char) : string;

// Zeichen am Ende lschen
function CutChar (const S : string; c : char) : string;

// Zeichen im String lschen
function DelChar (const S : string; c : char) : string;

// Leerzeichen gegen _ und / gegen % austauschen oder zurck
function ReplOptionChar (s : string; Back : boolean) : string;

// String auf feste Lnge bringen (krzen oder mit Leerzeichen auffllen)
function FixLength (const S : string;
                    len     : integer) : string;

// Anzahl von SubStr in S zhlen
function CountSubStr (const ASubStr,AString : string) : integer;

// Replace characters in "AText" from "Index1" to "Index2" with string "AToText"
function ReplaceChars (const AText,AToText : string; Index1,Index2 : cardinal) : string;

// Dateifilter-Index ermitteln (siehe TOpenDialog)
function GetFilterIndex(AFilter,AExtension : string) : integer;

{ ---------------------------------------------------------------- }
// Ausmaskieren der Erweiterung 
function GetExt (const Name : string) : string;

// Dateinamenerweiterung entfernen 
function DelExt (Name : string) : string;

// ndern einer Dateinamenerweiterung 
function NewExt (Name,Ext  : string) : string;

// Dateiname mit Laufwerk, Verzeichnis und Extension (falls nicht vorhanden) erweitern
function Erweiter (Pfad,Name,Ext : string) : string;

// Suffix zu Dateinamen hinzufgen
function AddNameSuffix (FName,Suffix,Ext : string) : string;

// Suffix zu Dateinamen mit Extension hinzufgen
function InsertNameSuffix (FName,Suffix : string) : string;

(* Suffix (Lnge: CharCount) von Dateinamen mit Extension entfernen *)
function RemoveNameSuffix (FName : string; CharCount : integer) : string;

// Pfad aus Dateinamen teilweise entfernen, wenn der Name zu lang wird
function StripPath (Name : String;
                    Len  : integer) : String;

// Bei Verzeichnisangaben Backslash ergnzen 
function SetDirName (Dir : string) : string;

// Exrahiere den bergeordneten Pfad 
function ExtractParentPath (Dir : string) : string;

// kein spezielles Verzeichnis (selbst und eins zurck) 
function NotSpecialDir (const Name : string) : boolean;

// enthlt vollstndigen Pfad 
function ContainsFullPath (const Name : string) : boolean;

// abs. Pfad in rel. Pfad umwandel und umgekehrt
function MakeRelativePath (const BaseName,DestName: string) : string;
function ExpandPath (const DestName: string) : string;

// prfe, ob alle Zeichen in Filename zulssig sind 
function CheckFilename (const Filename : string) : boolean;

// Ersetze ungltige Zeichen in Filename durch AChar
function ReplaceIllegalChars (const Filename : string; AChar : char) : string;

// identischen Pfadanteil am Anfang extrahieren
function ExtractSamePath(const Path1,Path2 : string) : string;

{ ---------------------------------------------------------------- }
// Routinen zur Auswertung einer Befehlszeile
// prfe, ob die ersten Zeichen einer Option mit dem Parameter bereinstimmen
function CompareOption (Param,Option : string) : boolean;

// Option vom Typ /option:value einlesen
function ReadOptionValue (var Param : string; Option : string) : boolean;

// Steuerdatei mit Startoptionen einlesen
function ReadOptionFile (FName : string) : string;

{ ---------------------------------------------------------------- }
(* Integer, Double oder String aus einem String s bis zum nchsten
   Begrenzer (Del) lesen
   s wird um den verarbeiteten Teil gekrzt *)
function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64;
                     var err : boolean) : int64; overload;

function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64) : int64; overload;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double;
                     var err : boolean) : double; overload;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double) : double; overload;

function ReadNxtStr (var s   : String;
                     Del     : char) : string; overload;

function ReadNxtStr (var s   : String;
                     Del     : char;
                     Default : string) : string; overload;

function ReadNxtWord (var s : string;
                      Dels  : TDelimiters) : string;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char;
                     var err : boolean) : char; overload;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char) : char; overload;

function ReadNxtQuotedStr (var s : string;
                           ADelim,AQuote : char) : string;

function MakeQuotedStr (const s : string; CheckChar : char) : string;

function CommaTextToLines (Commatext : string) : string;

// Position of char in string starting at "Offset", ignore parts between quotes
function PosChar (const s: string; AChar,AQuote : char; Offset: Cardinal): Integer;

{ ---------------------------------------------------------------- }
(* Testen, ob eine gltige Integer oder Double Zahl in einem String s
   bis zum nchsten Begrenzer (Del) steht *)
function TestNxtNumber (s       : String;
                        Del     : char ) : boolean;

// Wert auf Gltigkeit prfen - Result=true : x <> NoVal
function Valid (x,NoVal : double) : boolean;

// TFPoint-Struktur aus Koordinaten erzeugen
function FPoint (ax,ay : double) : TFPoint;

{ ---------------------------------------------------------------- }
// Dezimaltrenner setzen und lesen, Standard = Punkt 
procedure SetDecPtChar (Zeichen : char);
function GetDecPtChar : char;

// Tausendermarkierung, Standard = Leerzeichen 
procedure ThMarkChar (Zeichen : char);

// Integer-Zahl in String umsetzen 
function StrInt (x : int64;
                 n : integer) : string;

// Integer-Zahl in String mit fhrenden Nullen umsetzen 
function ZStrInt (x : int64;
                  n : integer) : string;

// Integer-Zahl als Hex-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToHexStr (Value,Digits,Group : cardinal) : string;

// Integer-Zahl als Dezimal-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToDecimal(Value,Digits,Group : cardinal) : string;

// Integer-Zahl als Oktal-Zahl, Digits = Mindestanzahl der Stellen,
//      Group = Gruppierung von Stellen
function IntToOctal (Value,Digits,Group : cardinal) : string;

// Integer-Zahl als Binr-Zahl, Digits = Mindestanzahl der Stellen
//      Group = Gruppierung von Stellen
function IntToBin (Value,Digits,Group : cardinal) : string;

// Konvertiere String als Hexwert (ohne $)
function HexStrToInt (S : string): cardinal;

// Konvertiere String als Oktalwert
function OctalStrToInt (S : string): cardinal;

// Konvertiere String als Binrwert
function BinStrToInt (S : string): cardinal;

// Double-Zahl in String umsetzen
function StrDbl (x   : double;
                 n,d : integer) : string;

// Double-Zahl in String mit fhrenden Nullen umsetzen
function ZStrDbl (x   : double;
                  n,d : integer) : string;

// Double-Zahl in String in Dreiergruppen umsetzen
function NStrDbl (x : double;
                  n,d : integer) : string;

// Bytes, je nach Gre in String umsetzen 
function SizeToStr (Bytes : int64) : string;

// String in Dateigre wandeln
function StrToSize (s : string; var size : int64) : boolean;

// Dateigre als Int64 aus Finddata
function GetFileSize (const FindData : TWin32FindData) : int64;

{ ---------------------------------------------------------------- }
// Dezimalkomma ersetzen
function ReplDecPt (s : string) : string;

// Dezimalpunkt einsetzen 
function SetDecPt (s : string) : string;

(* alle aktuellen Dezimaltrenner durch Punkt ersetzen *)
function ReplaceDecSeps (const s : string) : string;

(* alle Punkte durch aktuellen Dezimlatrenner ersetzen *)
function SetDecSeps (const s : string) : string;
{ ---------------------------------------------------------------- }
// Cardinal in Hi- oder Lo-Wort zerlegen 
function HiWord (n : cardinal) : word;

function LoWord (n : cardinal) : word;

// Cardinal aus Hi- und Lo-Wort zusammensetzen 
function JoinWord (HiWord,LoWord : word) : cardinal;

// Cardinal aus 4 Bytes zusammensetzen 
function JoinByte (HHByte,HLByte,LHByte,LLByte : word) : cardinal;

// Int64 aus Hi- und Lo-Cardinal zusammensetzen
function JoinCardinal (HiCard,LoCard : cardinal) : int64;

// Convert 16-bit oder 8-Bit number to hex string
function HexByte (b : byte) : string;

function HexWord (w : word) : string;

{ ---------------------------------------------------------------- }
// schnelles Sortieren einer Liste
procedure QuickSort (var SList;
                     Count,RSize : Integer;
                     Compare     : TCompFunction);

{ ---------------------------------------------------------------- }
// Prfe, ob Textdatei vorhanden
function ExistsFile (var f : TextFile) : boolean;

// Prfen, ob in einem Verzeichnis eine Datei existiert, die der Maske entspricht
function FileMatchesMask (Dir,Mask : string) : string;

// Datei kopieren
procedure FileCopy (srcfilename,destfilename : String); overload;
procedure FileCopy (srcfilename,destfilename : String; Attr : word); overload;

{ ---------------------------------------------------------------- }
// Dateien in einem Verzeichnis zhlen und Gesamtgre bestimmen
procedure CountFiles (const Base,Dir : string; IncludeSubDir : boolean;
                      var FileCount : cardinal; var FileSize : int64);
function DirFiles (const Directory : string; IncludeSubDir : boolean) : cardinal;
function DirSize (const Directory : string; IncludeSubDir : boolean) : int64;

(* Lsche ein Verzeichnis einschlielich aller Unterverzeichnisse und Dateien *)
procedure DeleteDirectory (const Base,Dir           : string;
                           var DCount,FCount,ECount : cardinal);
function DirDelete (const Directory : string) : boolean;

{ ---------------------------------------------------------------- }
// erweiterte Drucker-Angaben (alle Angaben in mm)
function GetPaperWidth (APrinter : TPrinter) : integer;
function GetPaperHeight (APrinter : TPrinter) : integer;
function GetLeftOffset (APrinter : TPrinter) : integer;
function GetTopOffset (APrinter : TPrinter) : integer;
function GetMaxWidth (APrinter : TPrinter) : integer;
function GetMaxHeight (APrinter : TPrinter) : integer;

// Duplex-Druck
function SupportsDuplex (APrinter : TPrinter) : Boolean;
function UsesDuplex (APrinter : TPrinter) : Boolean;
procedure SetToDuplex (APrinter : TPrinter);

{ ---------------------------------------------------------------- }
// Prfen, ob ein Fenster auf den Bildschirm passt
procedure CheckScreenBounds (Screen         : TScreen;
                             var ALeft,ATop : integer;
                             AWidth,AHeight : integer);

procedure FitToScreen (Screen : TScreen; Control : TControl);

{ ---------------------------------------------------------------- }
// MessageDlg in Bildschirmmitte (X<0) oder an Position X,Y
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons; Pos : TPoint) : integer; overload;
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons) : integer; overload;
function MessageDialog(const Msg: string; DlgType: TMsgDlgType;
                       Buttons: TMsgDlgButtons) : integer; overload;

function ConfirmDialog (const Title,Msg : string; Pos : TPoint) : boolean; overload;
function ConfirmDialog (const Title,Msg : string) : boolean; overload;
function ConfirmDialog (const Msg : string) : boolean; overload;

procedure InfoDialog (const Title,Msg : string; Pos : TPoint); overload;
procedure InfoDialog (const Title,Msg : string); overload;
procedure InfoDialog (const Msg : string); overload;

procedure ErrorDialog (const Title,Msg : string; Pos : TPoint); overload;
procedure ErrorDialog (const Title,Msg : string; x,y : integer); overload;
procedure ErrorDialog (const Title,Msg : string); overload;
procedure ErrorDialog (const Msg : string); overload;

{ ---------------------------------------------------------------- }
// get current cursor position
function CursorPos : TPoint; overload;
function CursorPos (Offset : TPoint): TPoint; overload;

// position of component
function TopLeftPos (AControl : TControl) : TPoint; overload;
function TopLeftPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function BottomLeftPos (AControl : TControl) : TPoint; overload;
function BottomLeftPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function TopRightPos (AControl : TControl) : TPoint; overload;
function TopRightPos (AControl : TControl; Offset : TPoint) : TPoint; overload;
function BottomRightPos (AControl : TControl) : TPoint; overload;
function BottomRightPos (AControl : TControl; Offset : TPoint) : TPoint; overload;

{ ---------------------------------------------------------------- }
// History list management
procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings; MaxCount : integer); overload;
procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings); overload;
procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer); overload;
procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings); overload;
procedure AddToHistory (History : TStrings; hs : string; MaxCount : integer); overload;
procedure AddToHistory (History : TStrings; hs : string); overload;
procedure RemoveFromHistory (History : TStrings; hs : string);

{ ---------------------------------------------------------------- }
// Read keys from ini file
function ReadStringFromIni (IniName,Section,Ident,Default : string) : string;
function ReadIntegerFromIni (IniName,Section,Ident : string; Default : integer) : integer;
function ReadBoolFromIni (IniName,Section,Ident : string; Default : boolean) : boolean;

{ ---------------------------------------------------------------- }
// Entferne alle Objekte einer String-Liste aus dem Speicher
procedure FreeListObjects(Liste : TStrings);

{ ---------------------------------------------------------------- }
// Listview-Index aus Caption ermitteln (wie IndexOf bei TListBox)
function GetListViewIndex (lv : TListView; ACaption : string): integer;

{ ---------------------------------------------------------------- }
// Tastaturpuffer lschen
function ClearKeyboardBuffer : Integer;

{ ---------------------------------------------------------------- }
implementation

uses StrUtils, WinSpool;

const
  rsHexError = '%s ist kein gltiger Hexwert';
  rsOctError = '%s ist kein gltiger Oktalwert';
  rsBinError = '%s ist kein gltiger Binrwert';
  rsErrCopy = 'Fehler beim Kopieren von %s';

{ ---------------------------------------------------------------- }
(* Umwandeln eines Zeichens in eines Grobuchstaben (auch Umlaute)
   OEM-Zeichen *)
function OEMUpCase (c : char) : char;
begin
  if c<#127 then Result:=upcase(c)
  else begin
    case c of
    '' : Result:='';
    '' : Result:='';
    '' : Result:='';
    '' : Result:='S';
      else Result:=c;
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Zeichens in eines Grobuchstaben (auch Umlaute)
   ANSI-Zeichen *)
function ANSIUpCase (c : char) : char;
begin
  if c<#127 then Result:=upcase(c)
  else begin
    case c of
    '' : Result:='';
    '' : Result:='';
    '' : Result:='';
    '' : Result:='S';
      else Result:=c;
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Strings in Grobuchstaben (auch Umlaute) *)
function OEMUpString (const S : string) : string;
var
  i  : integer;
  ns : string;
  c  : char;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  OEMUpString:=ns;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines ANSI-Strings in Grobuchstaben (auch Umlaute) *)
function ANSIUpString (const S : string) : string;
var
  i  : integer;
  ns : string;
  c  : char;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'';
      '' : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  ANSIUpString:=ns;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines Strings in Grobuchstaben (Umlaute in AE, OE, UE)
   zu Sortierzwecken *)
function OEMUpSort (const S : string) : string;
var
  i  : integer;
  ns : string;
  c  : char;
begin
  ns:='';
  for i:=1 to length(s) do begin
    c:=s[i];
    if c<#127 then ns:=ns+upcase(c)
    else begin
      case c of
      '','' : ns:=ns+'AE';
      '','' : ns:=ns+'OE';
      '','' : ns:=ns+'UE';
      ''     : ns:=ns+'SS';
        else ns:=ns+c;
        end;
      end;
    end;
  OEMUpSort:=ns;
  end;

{ ---------------------------------------------------------------- }
// Umwandeln eines Zeichens in Kleinbuchstaben
function LowCase(Ch: Char): Char;
begin
  Result:=Ch;
  if (Ch>='A') and (Ch<='Z') then inc(Result,Ord('a') - Ord('A'));
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines OEM-Strings nach ANSI *)
function StrToAnsi (const s : string) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToANSI[ord(Result[i])-128]
    else if Result[i]=#$15 then Result[i]:=#$A7;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Umwandeln eines ANSI-Strings nach OEM *)
function StrToOEM (const s : string) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToOEM[ord(Result[i])-128];
    end;
  end;

{------------------------------------------------------------------}
(* Substring-Pos. ohne Unterschreidung gro/klein *)
function TextPos (const Substr, S: string): Integer;
begin
  Result:=Pos (AnsiLowerCase(Substr),AnsiLowerCase(s));
  end;

function TextPosEx (const Substr, S: string; Offset : cardinal): Integer;
begin
  Result:=PosEx (AnsiLowerCase(Substr),AnsiLowerCase(s),Offset);
  end;


{ --------------------------------------------------------------- }
// Text auf Gro- oder Kleinschreibung ndern
function TextChangeCase(s: string; CaseMode : TTextChange) : string;
begin
  if CaseMode=tcLower then Result:=AnsiLowercase(s)
  else if CaseMode=tcUpper then Result:=AnsiUppercase(s)
  else Result:=s;
  end;

{ --------------------------------------------------------------- }
// Info-String aus zwei Zeilen erzeugen
function MakeInfoStr (s1,s2 : string) : string;
begin
  Result:=StringReplace(s1,'&','&&',[rfReplaceAll])+sLineBreak+
          StringReplace(s2,'&','&&',[rfReplaceAll]);
  end;

{------------------------------------------------------------------}
(* String mit Leerstellen erzeugen *)
function FillSpace (len : integer) : string;
var
  i : byte;
  s : string;
begin
  if len>255 then len:=255;
  s:='';
  for i:=1 to len do s:=s+' ';
  FillSpace:=s;
  end;

{------------------------------------------------------------------}
(* fhrende Leerstellen ergnzen *)
function AddSp (const S : string;
                len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=' '+Result;
  end;

{------------------------------------------------------------------}
(* Leerstellen am Ende ergnzen *)
function ExtSp (const S : string;
                len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=Result+' ';
  end;

{------------------------------------------------------------------}
(* fhrende Leerstellen und Tabs lschen *)
function RemSp (const S : string) : string;
var
  i : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) and ((Result[i]=Space) or (Result[i]=Tab)) do inc(i);
  delete(Result,1,pred(i));
  end;

{ --------------------------------------------------------------- }
(* Leerstellen am Ende lschen *)
function CutSp (const S : string) : string;
var
  i : integer;
begin
  Result:=s; i:=length(Result);
  while (i>0) and (Result[i]=Space) do dec(i);
  delete(Result,succ(i),length(Result));
  end;

{ --------------------------------------------------------------- }
(* alle Leerstellen im String lschen *)
function DelSp (const S : string) : string;
var
  i  : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) do begin
    if Result[i]=Space then delete(Result,i,1) else inc(i);
    end;
  end;

{ --------------------------------------------------------------- }
(* mehrfache Leerstellen im String lschen *)
function DelMulSp (const S : string) : string;
var
  i  : integer;
begin
  Result:=s;
  if length(Result)>0 then begin
    i:=1;
    while (i<length(Result)) do begin
      if (Result[i]=Space) and (Result[i+1]=Space) then delete(Result,i,1) else inc(i);
      end;
    if (Result[i]=Space) then delete(Result,i,1);
    end;
  end;

{ --------------------------------------------------------------- }
(* Steuerzeichen im String lschen *)
function DelCtrlChars(const S : string) : string;
var
  i  : integer;
begin
  Result:=s;
  if length(Result)>0 then begin
    i:=1;
    while (i<=length(Result)) do begin
      if Result[i]<#32 then delete(Result,i,1) else inc(i);
      end;
    end;
  end;

// Cr-Lf im String durch spez. Zeichen ersetzen
function CrLfToChar (s : string; c : char) : string;
begin
  Result:=DelCtrlChars(AnsiReplaceStr(s,#$0D,c));
  end;

// spez. Zeichen im String durch Cr-Lf ersetzen
function CharToCr (s : string; c : char) : string;
begin
  Result:=AnsiReplaceStr(s,c,#$0D);
  end;

{ --------------------------------------------------------------- }
(* erstes nichtleeres Zeichen suchen *)
function SeekNoSp (const s : string) : integer;
var
  i  : integer;
begin
  i:=1;
  while (i<=length(s)) and ((s[i]=Space) or (s[i]=Tab)) do inc(i);
  if i<=length(s) then SeekNoSp:=i else SeekNoSp:=0;
  end;

{ --------------------------------------------------------------- }
(* Position fr nchstes Leerzeichen oder Tab suchen,
   bei Stringende = Lnge+1 *)
function SeekSpace (const s : string) : integer;
var
  i  : integer;
begin
  if length(s)=0 then Result:=0
  else begin
    i:=1;
    while (i<=length(s)) and (s[i]<>Space) and (s[i]<>Tab) do inc(i);
    Result:=i;
    end;
  end;

{ --------------------------------------------------------------- }
(* String mit Zeichen erzeugen *)
function FillStr (c   : char;
                  len : integer) : string;
var
  i : byte;
  s : string;
begin
  if len>255 then len:=255;
  s:='';
  for i:=1 to len do s:=s+c;
  FillStr:=s;
  end;

{ --------------------------------------------------------------- }
(* fhrende Zeichen ergnzen *)
function AddChar (const S : string;
                  c       : char;
                  len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=c+Result;
  end;

{ --------------------------------------------------------------- }
(* Zeichen am Ende ergnzen *)
function ExtChar (const S : string;
                  c       : char;
                  len     : integer) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=succ(length(Result)) to len do Result:=Result+c;
  end;

{ --------------------------------------------------------------- }
// ein Zeichen durch ein anderes ersetzen
function ReplChars (const s         : string;
                   OldChar,NewChar : char) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do if Result[i]=OldChar then Result[i]:=NewChar;
  end;

{ --------------------------------------------------------------- }
// fhrende Zeichen lschen
function RemChar (const S : string; c : char) : string;
var
  i : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) and (Result[i]=c) do inc(i);
  delete(Result,1,pred(i));
  end;

{ --------------------------------------------------------------- }
// Zeichen am Ende lschen
function CutChar (const S : string; c : char) : string;
var
  i : integer;
begin
  Result:=s; i:=length(Result);
  while (i>0) and (Result[i]=c) do dec(i);
  delete(Result,succ(i),length(Result));
  end;

{ --------------------------------------------------------------- }
// Zeichen im String lschen
function DelChar (const S : string; c : char) : string;
var
  i  : integer;
begin
  Result:=s; i:=1;
  while (i<=length(Result)) do begin
    if Result[i]=c then delete(Result,i,1) else inc(i);
    end;
  end;

{ --------------------------------------------------------------- }
(* Leerzeichen gegen _ und / gegen % austauschen oder zurck *)
function ReplOptionChar (s : string; Back : boolean) : string;
var
  j : integer;
begin
  if Back then begin
    for j:=1 to length(s) do begin
      if s[j]='_' then s[j]:=' ';
      if s[j]='%' then s[j]:='/';
      end;
    end
  else begin
    for j:=1 to length(s) do begin
      if s[j]=' ' then s[j]:='_';
      if s[j]='/' then s[j]:='%';
      end;
    end;
  Result:=s;  
  end;

{ --------------------------------------------------------------- }
(* String auf feste Lnge bringen (krzen oder mit Leerzeichen auffllen *)
function FixLength (const S : string;
                    len : integer) : string;
begin
  if length(s)<len then FixLength:=ExtSp (S,len)
  else FixLength:=copy (s,1,len);
  end;

{ --------------------------------------------------------------- }
(* Anzahl von SubStr in S zhlen *)
function CountSubStr (const ASubStr,AString : string) : integer;
var
  n : integer;
begin
  Result:=0; n:=1;
  repeat
    n:=PosEx(ASubStr,AString,n+1);
    if n>0 then inc(Result);
    until n=0;
  end;

{ --------------------------------------------------------------- }
// Replace characters in "AText" from "Index1" to "Index2" with string "AToText"
function ReplaceChars (const AText,AToText : string; Index1,Index2 : cardinal) : string;
begin
  Result:=AText;
  Delete (Result,Index1,Index2-Index1+1);
  Insert (AToText,Result,Index1);
  end;

{ --------------------------------------------------------------- }
// Dateifilter-Index ermitteln (siehe TOpenDialog)
//
function GetFilterIndex(AFilter,AExtension : string) : integer;
var
  n : integer;
begin
  Result:=0; n:=0;
  repeat
    inc(n);
    ReadNxtStr(AFilter,'|');  // Beschreibung berlesen
    if AnsiContainsText(ReadNxtStr(AFilter,'|'),AExtension) then Result:=n;
    until (Result>0) or (length(AFilter)=0);
  if Result=0 then Result:=n;  // letztes Filter (*.*)
  end;

{ --------------------------------------------------------------- }
(* Ausmaskieren der Erweiterung (ohne Punkt) *)
function GetExt (const name : string) : string;
var
  i,j : integer;
begin
  Result:='';
  j:=length(Name);
  if j>0 then begin
    i:=j;
    while (i>0) and (Name[i]<>Punkt) and (not IsPathDelimiter(Name,i)) do dec(i);
    if (i>0) and (Name[i]=Punkt) then Result:=copy(name,i+1,j-i)
    end;
  end;

{ --------------------------------------------------------------- }
(* Dateinamenerweiterung entfernen (sucht letzten Punkt) *)
function DelExt (Name : string) : string;
var
  i,j : integer;
begin
  j:=length(Name);
  if j>0 then begin
    i:=j;
    while (i>0) and (Name[i]<>Punkt) and (not IsPathDelimiter(Name,i)) do dec(i);
    if (i>0) and (Name[i]=Punkt) then delete (Name,i,j-i+1);
    end;
  Result:=Name;
  end;

{ --------------------------------------------------------------- }
(* ndern einer Dateinamenerweiterung (sucht letzten Punkt - JR 9.8.05) *)
function NewExt (Name,Ext  : string) : string;
begin
  Name:=DelExt(Name);
  if (length(Ext)>0) and (Ext[1]=Punkt) then delete(Ext,1,1);
  if length(Ext)>0 then Result:=Name+Punkt+Ext
  else Result:=Name;
  end;

{ --------------------------------------------------------------- }
(* Dateiname mit Laufwerk, Verzeichnis und
   Extension (falls nicht vorhanden) erweitern
   Aug. 2007: Erweiter auch bei rel. Pfaden *)
function Erweiter (Pfad,Name,Ext  : string) : string;
var
  stemp : string;
begin
  if (length(Ext)>0) and (Ext[1]=Punkt) then delete(Ext,1,1);
  if (pos(Punkt,Name)=0) and (length(Ext)<>0) then Name:=Name+Punkt+ext;
  if (pos (DPunkt,Name)<>0) or (pos(BSlash,Name)=1)
     or (length(Pfad)=0) then stemp:=Name
  else if Pfad[length(Pfad)]=BSlash then stemp:=Pfad+Name
         else stemp:=Pfad+BSlash+Name;
  Result:=stemp;
  end;

{ ---------------------------------------------------------------- }
(* Suffix und Erweiterung zu Dateinamen hinzufgen *)
function AddNameSuffix (FName,Suffix,Ext : string) : string;
begin
  if (length(Ext)>0) and (Ext[1]<>Punkt) then Ext:=Punkt+Ext;
  Result:=DelExt(FName)+Suffix+Ext;
  end;

{ ---------------------------------------------------------------- }
(* Suffix zu Dateinamen mit Extension hinzufgen *)
function InsertNameSuffix (FName,Suffix : string) : string;
begin
  Result:=AddNameSuffix(FName,Suffix,GetExt(FName));
  end;

{ ---------------------------------------------------------------- }
(* Suffix (Lnge: CharCount) von Dateinamen mit Extension entfernen *)
function RemoveNameSuffix (FName : string; CharCount : integer) : string;
var
  s : string;
begin
  s:=DelExt(FName);
  Result:=AnsiLeftStr(s,length(s)-CharCount)+ExtractFileExt(FName);
  end;

{ ---------------------------------------------------------------- }
(* Pfad aus Dateinamen teilweise entfernen, wenn der Name zu lang wird
  siehe auch MinimizeName ind FileCtrl *)
function StripPath (Name : String;
                    Len  : integer) : String;
const
  Punkte = '...';
var
  i,j,nl : integer;
  ok     : boolean;
begin
  nl:=length(Name);
  if nl>=Len then begin
    i:=nl; ok:=true;
    if IsPathDelimiter(Name,i) then dec(i);
    while not IsPathDelimiter(Name,i) and (i>0) do dec(i);
    if i=0 then Name:=Punkte+copy(Name,nl-Len+3,nl)
    else begin
      dec(i); j:=i;
      repeat
        while not IsPathDelimiter(Name,j) and (j>0) do dec(j);
        dec(j);
        if j<0 then Name:=Punkte+copy(Name,nl-Len+3,nl)
        else begin
          ok:=nl-i+j+4<=Len;
          end;
        until ok or (j<0);
      if ok then begin
        inc(j,2);
        delete (Name,j,succ(i-j)); insert (Punkte,Name,j);
        end;
      end;
    end;
  Result:=Name;
  end;

{ ------------------------------------------------------------------- }
(* Extrahiere den bergeordneten Pfad      *)
function ExtractParentPath (Dir : string) : string;
var
  i : integer;
begin
  Dir:=ExcludeTrailingPathDelimiter(Dir);
  i:=length(Dir);
  while not IsPathDelimiter(Dir,i) and (i>0) do dec(i);
  if i<=1 then Result:=Dir
  else Result:=copy(Dir,1,i-1);
  end;

{ ------------------------------------------------------------------- }
(* Bei Verzeichnisangaben Backslash ergnzen *)
function SetDirName (Dir : string) : string;
begin
  if length(Dir)>0 then Result:=IncludeTrailingPathDelimiter(Dir)
  else Result:='';
//  if (length(Dir)=0) or (Dir[length(Dir)]=BSlash) then Result:=Dir
//  else Result:=Dir+BSlash;
  end;

{ ------------------------------------------------------------------- }
(* kein spezielles Verzeichnis (selbst und eins zurck) *)
function NotSpecialDir (const Name : string) : boolean;
begin
  Result:=(Name<>Punkt) and (Name<>'..');
  end;

{ ------------------------------------------------------------------- }
(* Name enthlt vollstndigen Pfad *)
function ContainsFullPath (const Name : string) : boolean;
begin
  if length(Name)>0 then Result:=(Name[1]=BSlash) or (pos(DPunkt,Name)>0)
  else Result:=false;
  end;

function MakeRelativePath (const BaseName,DestName: string) : string;
begin
  if AnsiStartsText(Basename,DestName) then Result:=ExtractRelativePath(BaseName,DestName)
  else Result:=DestName
  end;

function ExpandPath (const DestName: string) : string;
begin
  if ContainsFullPath(DestName) then Result:=DestName
  else Result:=ExpandFileName(Destname);
  end;

// identischen Teil des Pfads extrahieren  
function ExtractSamePath(const Path1,Path2 : string) : string;
var
  i     : integer;
  s1,s2 : string;
begin
  s1:=AnsiUpperCase(Path1);
  s2:=AnsiUpperCase(Path2);
  i:=1;
  while (i<=length(Path1)) and (i<=length(Path2))
    and (Path1[i]=Path2[i]) do inc(i);
  if (i<=length(Path1)) and (i<=length(Path2)) then Result:=copy(Path1,1,i-1)
  else Result:='';
  end;

{ ---------------------------------------------------------------- }
(* prfe, ob alle Zeichen in Filename zulssig sind *)
function CheckFilename (const Filename : string) : boolean;
var
  i : integer;
begin
  Result:=true; i:=1;
  if length(Filename)>0 then while Result and (i<=length(Filename)) do begin
    Result:=not (Filename[i] in IllegalFilenameChars);
    inc(i);
    end;
  end;

// Ersetze ungltige Zeichen in Filename durch AChar
function ReplaceIllegalChars (const Filename : string; AChar : char) : string;
var
  i : integer;
begin
  Result:=Filename;
  for i:=1 to length(Filename) do begin
    if (Result[i] in IllegalFilenameChars) then Result[i]:=AChar;
    end;
  end;

{ ---------------------------------------------------------------- }
// Routinen zur Auswertung einer Befehlszeile
// prfe, ob die ersten Zeichen einer Option mit dem Parameter bereinstimmen
function CompareOption (Param,Option : string) : boolean;
begin
  Result:=AnsiLowercase(Param)=copy(Option,1,length(Param));
  end;

// Option vom Typ option:value einlesen
function ReadOptionValue (var Param : string; Option : string) : boolean;
var
  i : integer;
begin
  Result:=false;
  i:=AnsiPos(DPunkt,Param);
  if i=0 then exit;
  if not CompareOption(copy(Param,1,i-1),Option) then exit;
  Delete(Param,1,i); Result:=true;
  end;

// Steuerdatei mit Startoptionen einlesen
function ReadOptionFile (FName : string) : string;
var
  f : TextFile;
  s : string;
begin
  Result:='';
  if FileExists(FName) then begin
    AssignFile(f,FName); Reset(f);
    while not Eof(f) do begin
      readln(f,s);
      s:=Trim(ReadNxtStr(s,';'));  // bis Kommentar ";" lesen
      if length(s)>0 then Result:=Result+'|'+s
      end;
    CloseFile(f);
    if length(Result)>0 then Delete(Result,1,1);
    end
  end;

{------------------------------------------------------------------}
(* Winkel-Wert konvertieren *)
function StrToDeg (s     : string;
                   var v : double) : boolean;
var
 t      : string;
 x,y    : double;
 j,k    : integer;
begin
  j:=0;
  k:=pos(',',s); if k>0 then s[k]:='.';
  if length(s)<>0 then begin
    s:=ReplDecPt(s);
    k:=pos(DPunkt,s);
    if k=0 then k:=pos('/',s);
    if k=0 then k:=pos('"',s);
    if k=0 then val(s,v,j)
    else begin (* Grad:Min *)
      val (copy(s,1,pred(k)),x,j); (* Grad *)
      if j=0 then begin
        t:=copy (s,succ(k),255);
        k:=pos(DPunkt,t);
        if k=0 then k:=pos('/',t);
        if k=0 then k:=pos(#$27,s);
        if k=0 then begin (* dez.Min *)
          if length(t)>0 then begin
            val (t,v,j);
            if (j=0) and (v<60.0) then begin
              if x>=0.0 then v:=x+v/60.0
              else v:=x-v/60.0;
              end
            else j:=1;
            end
          else v:=x;
          end
        else begin (* Sek. *)
          val (copy (t,1,pred(k)),y,j);  (* Min *)
          if (j=0) and (y<60.0) then begin
            t:=copy (t,succ(k),255);
            if length(t)>0 then begin
              val (t,v,j); (* Sek. *)
              if (j=0) and (v<60.0) then begin
                if x>=0.0 then v:=x+(y+v/60.0)/60.0
                else v:=x-(y-v/60.0)/60.0;
                end
              else j:=1;
              end
            else begin
              if x>=0.0 then v:=x+y/60.0
              else v:=x-y/60.0;
              end;
            end
          else j:=1;
          end;
        end;
      end;
    end
  else v:=0.0;
  StrToDeg:=j=0;
  end;

{------------------------------------------------------------------}
(* Winkel in String mit vorgebenem Format umsetzen
   Stellen: w : gesamt
            d : hinter Komma (einschl. Minuten und Sekunden
   Typ : m = fmDezGrad  : Grad,xxx
             fmDezMin   : Grad:Minuten,xx
             fmSekunden : Grad:Minuten:Sekunden,xxx *)
function DegToStr(v      : double;
                  Format : TWinkelFormat;
                  w,d    : integer) : string;
var
  s,t   : string;
  x,y,z : extended;
  i     : integer;
begin
  x:=int(abs(v)); y:=60.0*frac(abs(v));
  case Format of
  fmDezGrad : str(v:w:d,s);
  fmDezMin : begin (* Grad:Min *)
      str (y:(w-6):(d-2),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; x:=x+1.0;
        str (y:(w-6):(d-2),t);
        end;
      if v<0.0 then x:=-x;
      str(x:5:0,s); s:=s+DPunkt+t;
      end;
  fmSekunden : begin (* Grad:Min:Sek *)
      z:=int(y); y:=60.0*frac(y);
      str(y:(w-8):(d-4),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; z:=z+1.0;
        if z>=60.0 then begin
          z:=0.0; x:=x+1.0;
          end;
        str (y:(w-8):(d-4),t);
        end;
      str (z:2:0,s); t:=s+DPunkt+t;
      if v<0.0 then x:=-x;
      str(x:4:0,s); s:=s+':'+t;
      end;
    end;
  i:=pos('.',s); if i>0 then s[i]:=',';
  DegToStr:=s;
  end;

{ ------------------------------------------------------------------- }
(* Integer, Double oder String aus einem String s bis zum nchsten Tenner lesen
   s wird um den verarbeiteten Teil gekrzt *)
function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64;
                     var err : boolean) : int64;
var
  n    : int64;
  i,ic : integer;
begin
  i:=pos (Del,s);
  if i=0 then i:=succ(length(s));
  val(Trim(copy(s,1,pred(i))),n,ic);
  if ic=0 then ReadNxtInt:=n
  else begin
    ReadNxtInt:=Default;
    err:=true;
    end;
  delete(s,1,i);
  end;

function ReadNxtInt (var s   : String;
                     Del     : char;
                     Default : int64) : int64;
var
  err : boolean;
begin
  Result:=ReadNxtInt(s,Del,Default,err);
  end;

{ ------------------------------------------------------------------- }
function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double;
                     var err : boolean) : double;
var
  i,ic : integer;
  x    : double;
begin
  s:=RemSp(s); i:=pos (Del,s);
  if i=0 then i:=succ(length(s));
  val(SetDecPt(copy(s,1,pred(i))),x,ic);
  if ic=0 then Result:=x
  else begin
    Result:=Default;
    err:=true;
    end;
  delete(s,1,i);
  end;

function ReadNxtDbl (var s   : String;
                     Del     : char;
                     Default : double) : double;
var
  err : boolean;
begin
  Result:=ReadNxtDbl(s,Del,Default,err);
  end;

{ ------------------------------------------------------------------- }
function ReadNxtStr (var s   : String;
                     Del     : char) : string;
var
  i : integer;
begin
  if length(s)>0 then begin
    i:=pos (Del,s);
    if i=0 then i:=succ(length(s));
    Result:=copy(s,1,pred(i));
    delete(s,1,i);
    end
  else Result:='';
  end;

function ReadNxtStr (var s   : String;
                     Del     : char;
                     Default : string) : string;
var
  t : string;
begin
  t:=ReadNxtStr(s,Del);
  if length(t)=0 then Result:=Default else Result:=t;
  end;

{ ------------------------------------------------------------------- }
(* Wort bis zum nchsten Delimiter lesen *)
function ReadNxtWord (var s : string;
                      Dels  : TDelimiters) : string;
var
  i,n : integer;
begin
  n:=length(s);
  if n>0 then begin
    i:=1;
    while (i<=n) and not (s[i] in Dels) do inc(i);
    Result:=copy(s,1,pred(i));
    delete(s,1,pred(i));
    end
  else Result:='';
  end;

{ ------------------------------------------------------------------- }
(* ein Zeichen lesen oder num. Wert #nnn in Zeichen umwandeln *)
function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char;
                     var err : boolean) : char;
var
  n,i,ic : integer;
begin
  if (length(s)>0) then begin
    i:=pos (Del,s);
    if (s[1]=ChrID) then begin  (* # *)
      if i=0 then i:=succ(length(s));
      val(copy(s,2,pred(i)),n,ic);
      if ic=0 then ReadNxtChr:=chr(n)
      else begin
        ReadNxtChr:=Default; err:=true;
        end;
      end
    else ReadNxtChr:=s[1];
    delete(s,1,i);
    end
  else begin
    ReadNxtChr:=Default; err:=true;
    end;
  end;

function ReadNxtChr (var s   : String;
                     Del     : char;
                     Default : char): char;
var
  err : boolean;
begin
  Result:=ReadNxtChr(s,Del,Default,err);
  end;

function ReadNxtQuotedStr (var s : string; ADelim,AQuote : char) : string;
var
  n : integer;
  qs,fin : boolean;
begin
  s:=TrimLeft(s);
  if length(s)>0 then begin
    if s[1]=AQuote then begin
      n:=2; qs:=true; fin:=false;
      repeat
        if s[n]=AQuote then qs:=not qs
        else fin:=not qs and (s[n]=ADelim);
        if not fin then inc(n);
        until (n>length(s)) or fin;
      if n<=3 then result:=''
      else Result:=AnsiDequotedStr(TrimRight(copy(s,1,pred(n))),AQuote);
      end
    else begin
      n:=pos(ADelim,s);
      if n=0 then n:=succ(length(s));
      Result:=copy(s,1,pred(n));
      end;
    delete(s,1,n);
    end
  else Result:='';
  end;

function MakeQuotedStr (const s : string; CheckChar : char) : string;
begin
  if (length(s)=0) or ((CheckChar<>#0) and (Pos(CheckChar,s)=0)) then Result:=s
  else Result:=AnsiQuotedStr(s,Quote)
  end;

function CommaTextToLines (Commatext : string) : string;
begin
  Result:=ReadNxtQuotedStr(Commatext,',','"');
  while length(Commatext)>0 do
    Result:=Result+sLineBreak+ReadNxtQuotedStr(Commatext,',','"');
  end;

{ ------------------------------------------------------------------- }
// Position of char in string starting at "Offset", ignore parts between quotes
function PosChar (const s: string; AChar,AQuote : char; Offset: Cardinal): Integer;
var
  i,ls : integer;
  qs   : boolean;
begin
  i:=Offset;
  ls:=length(s);
  qs:=true;
  while i<=ls do begin
    if s[i]=AQuote then qs:=not qs
    else if qs and (s[i]=AChar) then begin
      Result:=i;
      exit;
      end;
    inc(i);
    end;
  Result:=0;
  end;

{ ------------------------------------------------------------------- }
(* Testen, ob gltige Integer oder Double Zahl in einem String s
   bis zum nchsten Begrenzer (Del) steht *)
function TestNxtNumber (s       : String;
                        Del     : char ) : boolean;
var
  n,i,ic : integer;
begin
  s:=RemSp(s); i:=pos (Del,s);
  if i=0 then i:=succ(length(s));
  val(copy(s,1,pred(i)),n,ic);
  TestNxtNumber:=ic=0;
  end;

{ ---------------------------------------------------------------- }
(* Dezimaltrenner auf Zeichen setzen, Standard = Punkt *)
procedure SetDecPtChar (Zeichen : char);
begin
  DezTr:=Zeichen;
  end;

{ ---------------------------------------------------------------- }
(* aktuellen Dezimaltrenner lesen *)
function GetDecPtChar : char;
begin
  GetDecPtChar:=DezTr;
  end;

{ ---------------------------------------------------------------- }
(* Tausendermarkierung, Standard = Leerzeichen *)
procedure ThMarkChar (Zeichen : char);
begin
  ThMark:=Zeichen;
  end;

{ ---------------------------------------------------------------- }
(* Integer-Zahl in String umsetzen *)
function StrInt (x : int64;
                 n : integer) : string;
(* n > 0 : Anzahl der Stellen
     = 0 : unformatiert
     < 0 : unformatiert mit Dreiergruppierung *)
var
  s   : string;
  i,j : integer;
begin
  s:=IntToStr(x);
  if n>length(s) then  s:=AddSp(s,n)
  else if n<0 then begin
    i:=succ(length(s));
    j:=i;
    while (j>4) do begin
      dec(j,3); insert (ThMark,s,j);
      end;
    end;
  StrInt:=s;
  end;

{ ---------------------------------------------------------------- }
(* Double-Zahl in String umsetzen *)
function StrDbl (x   : double;
                 n,d : integer) : string;
var
  s : string;
begin
  str (x:n:d,s);
  if DezTr<>DecPt then s:=ReplDecPt (s);
  StrDbl:=s;
  end;

{ ---------------------------------------------------------------- }
(* Integer-Zahl in String mit fhrenden Nullen umsetzen *)
function ZStrInt (x : int64;
                  n : integer) : string;
var
  s : string;
begin
  s:=IntToStr(abs(x));
  if x<0 then begin
    s:=AddChar(s,'0',n-1);
    s:='-'+s;
    end
  else s:=AddChar(s,'0',n);
  ZStrInt:=s;
  end;

{ ---------------------------------------------------------------- }
function GroupDigits (s : string; Group : integer) : string;
var
  i : integer;
begin
  if (Group>0) then begin
    i:=length(s);
    while (i>Group) do begin
      dec(i,Group); Insert(' ',s,i+1);
      end;
    end;
  Result:=s;
  end;

{ ---------------------------------------------------------------- }
function IntToHexStr (Value,Digits,Group : cardinal) : string;
begin
  Result:=GroupDigits(IntToHex(Value,Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToDecimal(Value,Digits,Group : cardinal) : string;
begin
  Result:=GroupDigits(ZStrInt(Value,Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToOctal (Value,Digits,Group : cardinal) : string;
begin
  Result:='';
  repeat
    Result:=chr((Value and 7)+48)+Result;
    Value:=Value div 8;
    until Value=0;
  Result:=GroupDigits(AddChar(Result,'0',Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function IntToBin (Value,Digits,Group : cardinal) : string;
begin
  Result:='';
  repeat
    Result:=chr((Value and 1)+48)+Result;
    Value:=Value shr 1;
    until Value=0;
  Result:=GroupDigits(AddChar(Result,'0',Digits),Group);
  end;

{ ---------------------------------------------------------------- }
function HexStrToInt (S : string): cardinal;
var
  minus : boolean;
begin
  Result:=0;
  if length(s)>0 then begin
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]<>'$') then s:='$'+s;
    s:=DelSp(s);
    try
      Result:=StrToInt(s);
      if minus then Result:=-Result;
    except
      raise EConvertError.Create(Format(rsHexError,[s]));
      exit;
      end;
    end;
  end;

function OctalStrToInt (S: string): cardinal;
var
  i,n : integer;
  minus : boolean;
begin
  Result:=0;
  if length(s)>0 then begin
    n:=0;
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]='%') then delete(s,1,1);     
    s:=DelSp(s);
    for i:=1 to length(s) do begin
      if (s[i]<#48) or (s[i]>#55) then begin
        raise EConvertError.Create(Format(rsOctError,[s]));
        exit;
        end
      else n:=8*n+ord(s[i])-48;
      end;
    if minus then n:=-n;
    Result:=n;
    end;
  end;

function BinStrToInt (S: string): cardinal;
var
  i,n : integer;
  minus : boolean;
begin
  Result:=0; 
  if length(s)>0 then begin
    n:=0;
    minus:=s[1]='-';
    if minus then delete(s,1,1);
    if (length(s)>0) and (s[1]='!') then delete(s,1,1);     
    s:=DelSp(s);
    for i:=1 to length(s) do begin
      if (s[i]<#48) or (s[i]>#49) then begin
        raise EConvertError.Create(Format(rsBinError,[s]));
        exit;
        end
      else n:=2*n+ord(s[i])-48;
      end;
    if minus then n:=-n;
    Result:=n;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Double-Zahl in String mit fhrenden Nullen umsetzen *)
function ZStrDbl (x   : double;
                  n,d : integer) : string;
var
  s : string;
  i : integer;
begin
  str (abs(x):n:d,s);
  if x<0 then if s[1]=Space then s[1]:='-' else s:='-'+s;
  for i:=1 to n do if s[i]=' ' then s[i]:='0';
  if DezTr<>DecPt then s:=ReplDecPt(s);
  ZStrDbl:=s;
  end;

{ ------------------------------------------------------------------ }
(* Double-Zahl in String in Dreiergruppen umsetzen*)
function NStrDbl (x : double;
                  n,d : integer) : string;
(* Format wie StrDbl *)
var
  s   : string;
  i,j : integer;
begin
  s:=StrDbl(x,n,d);
  i:=pos(DezTr,s);
  if i=0 then i:=succ(length(s));
  j:=i;
  while (j>3) do begin
    dec(j,3); insert (ThMark,s,j);
    inc(i);
    end;
  inc(i,3);
  while (i<length(s)) do begin
    insert(ThMark,s,i); inc(i);
    end;
  NStrDbl:=s;
  end;

{ ---------------------------------------------------------------- }
(* Bytes, je nach Gre in String umsetzen *)
function SizeToStr (Bytes : int64) : string;
var
  v : double;
  n : integer;
begin
  if Bytes<1024 then Result:=StrInt(Bytes,0)+' B'
  else begin
    v:=Bytes/1024;
    if v<1024 then begin
      if v<100 then n:=2 else n:=1;
      Result:=StrDbl(v,0,n)+' kB';
      end
    else begin
      v:=v/1024;
      if v<1024 then begin
        if v<100 then n:=2 else n:=1;
        Result:=StrDbl(v,0,n)+' MB'
        end
      else Result:=StrDbl(v/1024,0,2)+' GB';
      end;
    end;
  end;

// String in Dateigre wandeln
function StrToSize (s : string; var size : int64) : boolean;
const
 pfAnz = 4;
 pf : array[1..pfAnz] of char = ('k','M','G','T');
var
  f   : double;
  i,n : integer;
  c   : char;
begin
  Result:=false; size:=0;
  s:=Trim(s);
  if length(s)=0 then s:='0';
  if UpCase(s[length(s)])='B' then delete(s,length(s),1);
  s:=Trim(s);
  c:=s[length(s)];
  if (c>='0') then begin
    f:=1;
    if (c>'9') then begin
      n:=pos(' ',s);
      if n<=0 then Exit;
      i:=0;
      repeat
        f:=f*1024; inc(i);
        until (i>pfAnz) or (c=pf[i]);
      if i>pfAnz then Exit;
      s:=copy(s,1,n-1);
      end;
    try
      f:=StrToFloat(s)*f;
      size:=round(f);
    except
      Exit;
      end;
    Result:=true;
    end;
  end;
  
{ ---------------------------------------------------------------- }
// Dateigre als Int64 aus Finddata
function GetFileSize (const FindData : TWin32FindData) : int64;
begin
  with TInt64(Result) do begin
    Lo:=FindData.nFileSizeLow; Hi:=FindData.nFileSizeHigh;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Wert auf Gltigkeit prfen
   true : x <> NoVal *)
function Valid (x,NoVal : double) : boolean;
begin
  Valid:=abs(x-NoVal)>0.1;
  end;

{ ---------------------------------------------------------------- }
// TFPoint-Struktur aus Koordinaten erzeugen
function FPoint (ax,ay : double) : TFPoint;
begin
  with Result do begin
    x:=ax; y:=ay;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Dezimalpunkt durch vorgegebenen Trenner (SetDecPtChar) ersetzen
   Kompatibilittsfunktion, fr Neuentwicklungen "SetDecSeps" (s.u.) *)
function ReplDecPt (s : string) : string;
var
  i : byte;
begin
  i:=pos(DecPt,s);
  if i>0 then s[i]:=DezTr;
  ReplDecPt:=s;
  end;

(* aktuellen Dezimaltrenner durch Punkt ersetzen
   Kompatibilittsfunktion, fr Neuentwicklungen "ReplaceDecSeps" (s.u.) *)
function SetDecPt (s : string) : string;
var
  i : byte;
begin
  i:=pos (DezTr,s);
  if i>0 then s[i]:=DecPt;
  SetDecPt:=s;
  end;

(* alle aktuellen Dezimaltrenner durch Punkt ersetzen *)
function ReplaceDecSeps (const s : string) : string;
begin
  if DecimalSeparator<>DecPt then Result:=ReplChars(s,DecimalSeparator,DecPt);
  end;

(* alle Punkte durch aktuellen Dezimlatrenner ersetzen *)
function SetDecSeps (const s : string) : string;
begin
  if DecimalSeparator<>DecPt then Result:=ReplChars(s,DecPt,DecimalSeparator);
  end;

{ ---------------------------------------------------------------- }
(* Cardinal in Hi- oder Lo-Wort zerlegen *)
function HiWord (n : cardinal) : word;
var
  Long : TLongWord absolute n;
begin
  HiWord:=Long.Hi;
  end;

function LoWord (n : cardinal) : word;
var
  Long : TLongWord absolute n;
begin
  LoWord:=Long.Lo;
  end;

(* Cardinal aus Hi- und Lo-Wort zusammensetzen *)
function JoinWord (HiWord,LoWord : word) : cardinal;
var
  Long : TLongWord;
begin
  with Long do begin
    Hi:=HiWord; Lo:=LoWord;
    JoinWord:=LongWord;
    end;
  end;

(* Cardinal aus 4 Bytes zusammensetzen *)
function JoinByte (HHByte,HLByte,LHByte,LLByte : word) : cardinal;
var
  Long : TLongWord;
begin
  with Long do begin
    LoL:=LLbyte; LoH:=LHByte;
    HiL:=HLByte; HiH:=HHByte;
    JoinByte:=LongWord;
    end;
  end;

(* Int64 aus Hi- und Lo-Cardinal zusammensetzen *)
function JoinCardinal (HiCard,LoCard : cardinal) : int64;
var
  Long : TInt64;
begin
  with Long do begin
    Hi:=HiCard; Lo:=LoCard;
    Result:=AsInt64;
    end;
  end;

{ ---------------------------------------------------------------- }
// Convert 16-bit oder 8-Bit number to hex string
function HexByte (b : byte) : string;
const
  HexTab : array [0..15] of char = '0123456789ABCDEF';
begin
  HexByte:=HexTab[b shr 4] + HexTab[b and $f];
  end;

function HexWord (w : word) : string;
begin
  HexWord:=HexByte(hi(w))+HexByte(lo(w));
  end;

//-----------------------------------------------------------------------------
{ QUICKSORT sorts elements in the array "SList" of any record type of
  size "RSize" (in bytes) with indices between 0 and "Count"-1 (both inclusive).
  Note that the QuickSort procedure provides only an "interface" to the program.
  The actual processing takes place in the "Sort" procedure, which executes
  itself recursively.
  "Compare" is a callback function of type "TCompFunction" (see begin of unit).
  Algorithm based on Borland pascal sample "qsort.pas",
  modified J. Rathlev, Nov. 2001 (rathlev@physik.uni-kiel.de) }
//-----------------------------------------------------------------------------
type
  TByteDynArray = array of byte;

procedure QuickSort (var SList;
                     Count,RSize : Integer;
                     Compare     : TCompFunction);
var
  {the  following variables are global to "Sort" to spare memory in recursive calls }
  tmp,ns  : TByteDynArray;
  pl      : PByteArray;
  n,m     : integer;

  procedure Sort(nl,nr: Integer);
  var
    i,j   : integer;
  begin
    n:=RSize*((nl+nr) div 2);
    move (pl^[n],ns[0],RSize);
    i:=nl; j:=nr;
    repeat
      while Compare(pl^[RSize*i],ns[0]) do inc(i);
      while Compare(ns[0],pl^[RSize*j]) do dec(j);
      if i<=j then begin
        n:=RSize*i; m:=RSize*j;
        move (pl^[n],tmp[0],RSize);
        move (pl^[m],pl^[n],RSize);
        move (tmp[0],pl^[m],RSize);
        inc(i); dec(j);
        end;
      until i>j;
    if nl<j then Sort(nl,j);
    if i<nr then Sort(i,nr);
    end;

begin {QuickSort};
  SetLength(Tmp,RSize); SetLength(ns,RSize);
  pl:=@SList;
  Sort(0,Count-1);
  Tmp:=nil; ns:=nil;
  end;

{ ---------------------------------------------------------------- }
(* Diese Funktion prft, ob eine Text-Datei vorhanden ist und ffnet sie *)
function ExistsFile (var f : TextFile) : boolean;
begin
  (*$i-*) reset (f) (*$i+*);
  Result:=ioresult=0;
  end;

//-----------------------------------------------------------------------------
// Prfen, ob in einem Verzeichnis eine Datei existiert, die der Maske entspricht
function FileMatchesMask (Dir,Mask : string) : string;
var
  FInfo      : TSearchRec;
begin
  Result:='';
  if FindFirst (SetDirName(Dir)+Mask,faArchive,FInfo)=0 then Result:=SetDirName(Dir)+FInfo.Name;
  Sysutils.FindClose (FInfo);
  end;

{ ------------------------------------------------------------------- }
(* Datei kopieren mit Attribut und Datum - verwendet CopyFile aus WindowsAPI *)
procedure FileCopy (srcfilename,destfilename : String);
var
  ok : boolean;
begin
  if FileExists (srcfilename) and (length(destfilename)>0) then begin
    if FileExists(destfilename) then ok:=FileSetAttr(destfilename,0)=0
    else ok:=true;
    if ok then begin
      if not CopyFile(pchar(srcfilename),pchar(destfilename),false) then begin
        raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
        end;
      end
    else begin
      raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
      end;
    end;
  end;

(* Datei kopieren mit Datum und neuem Attribut *)
procedure FileCopy (srcfilename,destfilename : String; Attr : word);
begin
  try
    FileCopy(srcfilename,destfilename);
    FileSetAttr(destfilename,Attr);
  except
    raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
    end;
  end;

{ ---------------------------------------------------------------- }
// Dateien in einem Verzeichnis zhlen und Gesamtgre bestimmen
procedure CountFiles (const Base,Dir : string; IncludeSubDir : boolean;
                      var FileCount : cardinal; var FileSize : int64);
var
  DirInfo    : TSearchRec;
  Findresult : integer;
  sd         : string;
begin
  sd:=SetDirName(SetDirName(Base)+Dir);
  FindResult:=FindFirst (sd+'*.*',faAnyFile,DirInfo);
  while FindResult=0 do with DirInfo do begin
    if IncludeSubDir and NotSpecialDir(Name) and ((Attr and faDirectory)<>0) then
      CountFiles (Base,Erweiter(Dir,DirInfo.Name,''),IncludeSubDir,FileCount,FileSize);
    FindResult:=FindNext(DirInfo);
    end;
  FindClose(DirInfo);
  FindResult:=FindFirst (sd+'*.*',faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
  while FindResult=0 do with DirInfo do begin
    if NotSpecialDir(Name) then begin
      inc(FileCount); inc(FileSize,GetFileSize(DirInfo.FindData));
      end;
    FindResult:=FindNext(DirInfo);
    end;
  FindClose(DirInfo);
  end;

function DirFiles (const Directory : string; IncludeSubDir : boolean) : cardinal;
var
  n : int64;
begin
  Result:=0; n:=0;
  CountFiles(Directory,'',IncludeSubDir,Result,n);
  end;

function DirSize (const Directory : string; IncludeSubDir : boolean) : int64;
var
  n : cardinal;
begin
  Result:=0; n:=0;
  CountFiles(Directory,'',IncludeSubDir,n,Result);
  end;

(* Lsche ein Verzeichnis einschlielich aller Unterverzeichnisse und Dateien *)
procedure DeleteDirectory (const Base,Dir           : string;
                           var DCount,FCount,ECount : cardinal);
var
  DirInfo    : TSearchRec;
  fc,dc,
  Findresult : integer;
  s,sd       : string;
begin
  if length(Dir)>0 then sd:=IncludeTrailingPathDelimiter(Base)+Dir else sd:=Base;
  if DirectoryExists(sd) then begin
    FindResult:=FindFirst (Erweiter(sd,'*','*'),faAnyFile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) and ((Attr and faDirectory)<>0) then
        DeleteDirectory(Base,Erweiter(Dir,DirInfo.Name,''),DCount,FCount,ECount);
      FindResult:=FindNext(DirInfo);
      end;
    FindClose(DirInfo);
    fc:=0; dc:=0;
    FindResult:=FindFirst (Erweiter(sd,'*','*'),faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then begin
        inc(fc);
        (* Dateien lschen *)
        s:=SetDirName(sd)+Name;
        FileSetAttr(s,faArchive);
        if DeleteFile(s) then begin
          inc(FCount); inc(dc);
          end
        else inc(ECount);  // Fehler
        end;
      FindResult:=FindNext(DirInfo);
      end;
    FindClose(DirInfo);
    if (fc=dc) then begin   // Verzeichnis leer ==> lschen
      FileSetAttr(sd,0);    // Attribute zum Lschen entfernen
      if RemoveDir(sd) then inc(DCount) else inc(ECount);
      end;
    end;
  end;

function DirDelete (const Directory : string) : boolean;
var
  fc,dc,ec : cardinal;
begin
  fc:=0; dc:=0; ec:=0;
  DeleteDirectory(Directory,'',dc,fc,ec);
  Result:=ec=0;
  end;

{ ------------------------------------------------------------------- }
(* Erweiterte Druckerangaben *)
(* tatschliche Papierbreite in mm *)
function GetPaperWidth (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALWIDTH)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* tatschliche Papierhhe in mm *)
function GetPaperHeight (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALHEIGHT)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nichtdruckbarer Bereich am linken Rand in mm *)
function GetLeftOffset (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALOFFSETX)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nichtdruckbarer Bereich am oberen Rand in mm *)
function GetTopOffset (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result:=round(25.4*GetDeviceCaps(Handle,PHYSICALOFFSETY)/GetDeviceCaps(Handle,LOGPIXELSX));
  end;

(* nutzbare Papierbreite in mm *)
function GetMaxWidth (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result := GetDeviceCaps(Handle, HorzSize);
  end;

(* nutzbare Papierhhe in mm *)
function GetMaxHeight (APrinter : TPrinter) : integer;
begin
  with APrinter do
    Result := GetDeviceCaps(Handle, VertSize);
  Sleep(100);
  end;

// prfe, ob Duplex-Druck untersttzt wird
function SupportsDuplex (APrinter : TPrinter) : Boolean;
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
begin
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  Result:=WinSpool.DeviceCapabilities(Device,Port,DC_DUPLEX,nil,nil)<>0;
  end;

// prfe,ob Duplex-Druck eingestellt ist
function UsesDuplex (APrinter : TPrinter) : Boolean;
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
  pDevmode: PDeviceMode;
begin
  Result:=false;
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  if hDevmode<>0 then begin
     // lock it to get pointer to DEVMODE record
    pDevMode:=GlobalLock(hDevmode);
    if pDevmode<>nil then
      try
        Result:=pDevmode^.dmDuplex<>DMDUP_SIMPLEX;
      finally
        // unlock devmode handle.
        GlobalUnlock(hDevmode);
      end;
    end;
  end;

// Schalte Drucker auf Duplex
procedure SetToDuplex (APrinter : TPrinter);
var
  Device,Driver,Port : array[0..255] of Char;
  hDevMode: THandle;
  pDevmode: PDeviceMode;
begin
  APrinter.GetPrinter(Device,Driver,Port,hDevmode);
  if (hDevmode<>0) and (WinSpool.DeviceCapabilities(Device,Port,DC_DUPLEX,nil,nil)<>0) then begin
     // lock it to get pointer to DEVMODE record
    pDevMode:=GlobalLock(hDevmode);
    if pDevmode<>nil then
      try
        with pDevmode^ do begin
          dmDuplex:=DMDUP_VERTICAL;
          dmFields:=dmFields or DM_DUPLEX;
          end;
      finally
        // unlock devmode handle.
        GlobalUnlock(hDevmode);
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
(* Prfen, ob ein Fenster auf den Bildschirm passt, bei Bedarf
   Left, und Top anpassen *)
procedure CheckScreenBounds (Screen         : TScreen;
                             var ALeft,ATop : integer;
                             AWidth,AHeight : integer);
begin
  with Screen do begin
    if ALeft<0 then ALeft:=20;
    if ATop<0 then ATop:=20;
    if ALeft+AWidth>Width then ALeft:=Width-AWidth-20;
    if ATop+AHeight>Height then ATop:=Height-AHeight-20;
    end;
  end;

procedure FitToScreen (Screen : TScreen; Control : TControl);
var
  il,it : integer;
begin
  with Control do begin
    il:=Left; it:=Top;
    CheckScreenBounds (Screen,il,it,Width,Height);
    Left:=il; Top:=it;
    end;
  end;

{ ---------------------------------------------------------------- }
// neuer Message-Dialog mit Positionsprfung
function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons; Pos : TPoint) : integer;
begin
  with CreateMessageDialog(Msg, DlgType, Buttons) do
    try
      with Pos do begin
        if (Y < 0) and (X < 0) then Position:=poScreenCenter
        else begin
//          if X<0 then X:=Left;
//          if Y<0 then Y:=Top;
          CheckScreenBounds(Screen,x,y,Width,Height);
          Left:=x; Top:=y;
          end;
        end;
      if length(Title)>0 then Caption:=Title;
      FormStyle:=fsStayOnTop;
      Result:=ShowModal;
    finally
      Free;
    end;
  end;

function MessageDialog(const Title,Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons) : integer;
begin
  Result:=MessageDialog(Title,Msg,DlgType,Buttons,CenterPos);
  end;

function MessageDialog(const Msg: string; DlgType: TMsgDlgType;
  Buttons: TMsgDlgButtons) : integer;
begin
  Result:=MessageDialog('',Msg,DlgType,Buttons,CenterPos);
  end;

{ ---------------------------------------------------------------- }
// Besttigung in Bildschirmmitte (X<0) oder an Position X,Y
function ConfirmDialog (const Title,Msg  : string; Pos : TPoint) : boolean;
begin
  Result:=MessageDialog (Title,Msg,mtConfirmation,[mbYes,mbNo],Pos)=mrYes;
  end;

// Besttigung in Bildschirmmitte
function ConfirmDialog (const Title,Msg  : string) : boolean;
begin
  Result:=ConfirmDialog(Title,Msg,CenterPos);
  end;

function ConfirmDialog (const Msg  : string) : boolean;
begin
  Result:=ConfirmDialog('',Msg,CenterPos);
  end;

// Information an Position ausgeben
procedure InfoDialog (const Title,Msg : string; Pos : TPoint);
begin
  MessageDialog (Title,Msg,mtInformation,[mbOK],Pos);
  end;

// Information in Bildschirmmitte ausgeben
procedure InfoDialog (const Title,Msg :string);
begin
  InfoDialog(Title,Msg,CenterPos);
  end;

procedure InfoDialog (const Msg :string);
begin
  InfoDialog('',Msg,CenterPos);
  end;

// Fehlermeldung an Position ausgeben
procedure ErrorDialog (const Title,Msg : string; Pos : TPoint);
begin
  MessageDialog (Title,Msg,mtError,[mbOK],Pos);
  end;

procedure ErrorDialog (const Title,Msg : string; x,y : integer);
begin
  MessageDialog (Title,Msg,mtError,[mbOK],Point(x,y));
  end;

// Fehlermeldung in Bildschirmmitte ausgeben
procedure ErrorDialog (const Title,Msg : string);
begin
  ErrorDialog(Title,Msg,CenterPos);
  end;

procedure ErrorDialog (const Msg : string);
begin
  ErrorDialog('',Msg,CenterPos);
  end;

{ ------------------------------------------------------------------- }
// get current cursor position
function CursorPos : TPoint;
var
  pos : TPoint;
begin
  GetCursorPos(pos);
  Result:=pos;
  end;

{ ------------------------------------------------------------------- }
// Add offset to point
function AddOffsetPos(Pos1,Pos2 : TPoint) : TPoint;
begin
  with Result do begin
    x:=Pos1.x+Pos2.x; y:=Pos1.y+Pos2.y;
    end;
  end;

// get current cursor position, add Offset
function CursorPos (Offset : TPoint): TPoint;
var
  pos : TPoint;
begin
  GetCursorPos(pos);
  Result:=AddOffsetPos(pos,Offset);
  end;

{ ------------------------------------------------------------------- }
// position of component
function TopLeftPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left,Top))
  else Result:=Point(Left,Top);
  Result:=AddOffsetPos(Result,Offset);
  end;

function TopLeftPos (AControl : TControl) : TPoint;
begin
  TopLeftPos (AControl,Point(0,0));
  end;

function BottomLeftPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left,Top+Height))
  else Result:=Point(Left,Top+Height);
  Result:=AddOffsetPos(Result,Offset);
  end;

function BottomLeftPos (AControl : TControl) : TPoint;
begin
  BottomLeftPos (AControl,Point(0,0));
  end;

function TopRightPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left+Width,Top))
  else Result:=Point(Left+Width,Top);
  Result:=AddOffsetPos(Result,Offset);
  end;

function TopRightPos (AControl : TControl) : TPoint;
begin
  TopRightPos (AControl,Point(0,0));
  end;

function BottomRightPos (AControl : TControl; Offset : TPoint) : TPoint;
begin
  with AControl do if assigned(Parent) then Result:=Parent.ClientToScreen(Point(Left+Width,Top+Height))
  else Result:=Point(Left+Width,Top+Height);
  Result:=AddOffsetPos(Result,Offset);
  end;

function BottomRightPos (AControl : TControl) : TPoint;
begin
  BottomRightPos (AControl,Point(0,0));
  end;

{ ------------------------------------------------------------------- }
// History list management
const
  iniHist = 'History';

procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings; MaxCount : integer);
var
  i : integer;
  s : string;
begin
  with IniFile do begin
    if SectionExists(Section) then begin
      if length(Ident)=0 then Ident:=iniHist;
      History.Clear;
      for i:=0 to MaxCount-1 do begin
        s:=ReadString(Section,Ident+IntToStr(i),'');
        if s<>'' then History.Add(s);
        end;
      end;
    end;
  end;

procedure LoadHistory (IniFile : TIniFile; Section,Ident : string;
                       History : TStrings);
begin
  LoadHistory(IniFile,Section,Ident,History,MaxHist);
  end;

procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings; MaxCount : integer);
var
  i,n : integer;
begin
  with IniFile do begin
    if length(Ident)=0 then Ident:=iniHist;
    if Erase then EraseSection (Section);
    with History do begin
      if Count>MaxCount then n:=MaxCount else n:=Count;
      for i:=0 to n-1 do WriteString(Section,Ident+IntToStr(i),Strings[i]);
      end;
    end;
  end;

procedure SaveHistory (IniFile : TIniFile; Section,Ident : string;
                       Erase : boolean; History : TStrings);
begin
  SaveHistory(IniFile,Section,Ident,Erase,History,MaxHist);
  end;

// move or add item "hs" to begin of history list
procedure AddToHistory (History : TStrings; hs : string; MaxCount : integer);
var
  n : integer;
begin
  if length(hs)>0 then with History do begin
    n:=IndexOf(hs);
    if n<0 then begin
      if Count>=MaxCount then Delete (Count-1);
      Insert (0,hs);
      end
    else Move (n,0);
    end;
  end;

procedure AddToHistory (History : TStrings; hs : string);
begin
  AddToHistory (History,hs,MaxHist);
  end;

procedure RemoveFromHistory (History : TStrings; hs : string);
var
  n : integer;
begin
  if length(hs)>0 then with History do begin
    n:=IndexOf(hs);
    if n>=0 then Delete (n);
    end;
  end;

//-----------------------------------------------------------------------------
function ReadStringFromIni (IniName,Section,Ident,Default : string) : string;
begin
  with TIniFile.Create(IniName) do begin
    Result:=ReadString(Section,Ident,Default);
    Free;
    end;
  end;

function ReadIntegerFromIni (IniName,Section,Ident : string; Default : integer) : integer;
begin
  with TIniFile.Create(IniName) do begin
    Result:=ReadInteger(Section,Ident,Default);
    Free;
    end;
  end;

function ReadBoolFromIni (IniName,Section,Ident : string; Default : boolean) : boolean;
begin
  with TIniFile.Create(IniName) do begin
    Result:=ReadBool(Section,Ident,Default);
    Free;
    end;
  end;

//-----------------------------------------------------------------------------
procedure FreeListObjects (Liste : TStrings);
var
  i : integer;
begin
  with Liste do begin
    for i:=0 to Count-1 do if assigned(Objects[i]) then begin
      Objects[i].Free; Objects[i]:=nil;
      end;
    end;
  end;


//-----------------------------------------------------------------------------
// Listview-Index aus Caption ermitteln (wie IndexOf bei TListBox)
function GetListViewIndex (lv : TListView; ACaption : string): integer;
begin
  with lv.Items do for Result:=0 to Count-1 do
    if AnsiSameText(Item[Result].Caption,ACaption) then Exit;
  Result:=-1;
  end;

//-----------------------------------------------------------------------------
// Tastaturpuffer lschen
function ClearKeyboardBuffer : Integer;
var
   Msg: TMsg;
begin
  Result := 0;
  while PeekMessage(Msg,0,WM_KEYFIRST,WM_KEYLAST,PM_REMOVE) do inc(Result);
  end;

{ ------------------------------------------------------------------- }
(* Verzgerungszeit unter Verwendung der Windows Waitable Timer Objects
   Wait: warten bis Zeit in ms abgelaufen ist (min. Zeit 100 ns) *)

constructor TWaitTimer.Create;
begin
  inherited Create;
  FTimer:=CreateWaitableTimer(nil,true,'WaitableTimer');
  end;

destructor TWaitTimer.Destroy;
begin
  CloseHandle(FTimer);
  end;

procedure TWaitTimer.Wait (DelayTime : double);
var
  d : TLargeInteger;
begin
  d:=-round(10000*DelayTime);
  SetWaitableTimer(FTimer,d,0,nil,nil,false);
  repeat until WaitForSingleObject(FTimer,INFINITE)=WAIT_OBJECT_0;
  end;

{ ------------------------------------------------------------------- }
// Anzeige eines Hinweisefenster (THintWindow), das nach einstellbarer Zeit (Delay)
// automatisch verschwindet
constructor TTimerHint.Create (AOwner: TComponent; Delay : integer);
begin
  inherited Create(AOwner);
  FTimer:=TTimer.Create(AOwner);
  with FTimer do begin
    Interval:=Delay;
    Enabled:=false;
    OnTimer:=Terminate;
    end;
  FOnTerminate:=nil;
  end;

destructor TTimerHint.Destroy;
begin
  FTimer.Free;
  inherited Destroy;
  end;

procedure TTimerHint.ShowHint (r : TRect; AHint : string);
begin
  ActivateHint(r,AHint);
  FTimer.Enabled:=true;
  end;

procedure TTimerHint.Terminate (Sender : TObject);
begin
  FTimer.Enabled:=false;
  ReleaseHandle;
  if assigned(FOnTerminate) then FOnTerminate(self);
  end;

procedure TTimerHint.HideHint;
begin
  FTimer.Enabled:=false;
  ReleaseHandle;
  end;

{ ------------------------------------------------------------------- }
begin
  DezTr:=DecimalSeparator;
end.

