(* Delphi 7 - Unit
   Extended media functions

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Apr. 2002
   Vers. 2 - Dec. 2005 - with volume control
   *)

unit Sound;

interface

uses Windows, MPlayer, mmSystem;

type
  TSoundFlag = (sfAsync,sfSync,sfLoop,sfNoStop,sfNoDefault);
  TSoundFlags = set of TSoundFlag;

const
  SoundID = 'WAVE';

var
  SoundAvailable : boolean;

function PlaySoundFromFile (Filename : string;
                            Flags    : TSoundFlags) : boolean;

function PlaySoundFromMemory (ResID : string;
                              Flags : TSoundFlags) : boolean;
                            
procedure StopSound;
                            
{ Volume: 0 - 1000 }
function GetMPVolume(MP: TMediaPlayer): Integer;
procedure SetMPVolume(MP: TMediaPlayer; Volume: Integer);

implementation

function GetSoundFlags (Flags : TSoundFlags) : cardinal;
begin
  Result:=0;
  if sfAsync in Flags then Result:=Result or SND_ASYNC;
  if sfSync in Flags then Result:=Result or SND_SYNC;
  if sfLoop in Flags then Result:=Result or SND_LOOP;
  if sfNoStop in Flags then Result:=Result or SND_NOSTOP;
  if sfNoDefault in Flags then Result:=Result or SND_NODEFAULT;
  end;

function PlaySoundFromFile (Filename : string;
                            Flags    : TSoundFlags) : boolean;
begin
  if SoundAvailable then Result:=sndPlaySound(pchar(Filename),GetSoundFlags(Flags))
  else Result:=false;
  end;

function PlaySoundFromMemory (ResID : string;
                              Flags : TSoundFlags) : boolean;
var
  Res,rh : THandle;
begin
  Result:=false;
  if SoundAvailable then begin
    Res:=FindResource(HInstance,pchar(ResID),SoundID);
    if Res<>0 then begin
      rh:=LoadResource(HInstance,Res);
      if rh<>0 then begin
        Result:=sndPlaySound(LockResource(rh),GetSoundFlags(Flags) or SND_MEMORY);
        UnlockResource(rh);
        FreeResource(rh);
        end;
      end;
    end;
  end;

procedure StopSound;
begin
  if SoundAvailable then sndPlaySound(nil,0);
  end;

const
  MCI_SETAUDIO = $0873;
  MCI_DGV_SETAUDIO_VOLUME = $4002;
  MCI_DGV_SETAUDIO_ITEM = $00800000;
  MCI_DGV_SETAUDIO_VALUE = $01000000;
  MCI_DGV_STATUS_VOLUME = $4019;

type
  TMCI_DGV_SETAUDIO_PARMS = record
    dwCallback : DWORD;
    dwItem     : DWORD;
    dwValue    : DWORD;
    dwOver     : DWORD;
    lpstrAlgorithm : PChar;
    lpstrQuality   : PChar;
  end;

type
  TMCI_STATUS_PARMS = record
    dwCallback : DWORD;
    dwReturn   : DWORD;
    dwItem     : DWORD;
    dwTrack    : DWORD;
  end;

procedure SetMPVolume(MP: TMediaPlayer; Volume: Integer);
  { Volume: 0 - 1000 }
var
  AudioParms : TMCI_DGV_SETAUDIO_PARMS;
begin
  { Volume: 0 - 1000 }
  with AudioParms do begin
    dwCallback := 0;
    dwItem := MCI_DGV_SETAUDIO_VOLUME;
    dwValue := Volume;
    dwOver := 0;
    lpstrAlgorithm := nil;
    lpstrQuality := nil;
    end;
  mciSendCommand(MP.DeviceID, MCI_SETAUDIO,
    MCI_DGV_SETAUDIO_VALUE or MCI_DGV_SETAUDIO_ITEM, Cardinal(@AudioParms));
  end;

function GetMPVolume(MP: TMediaPlayer): Integer;
var
  StatusParms : MCI_STATUS_PARMS;
begin
  with StatusParms do begin
    dwCallback := 0;
    dwItem := MCI_DGV_STATUS_VOLUME;
    end;
  mciSendCommand(MP.DeviceID, MCI_STATUS, MCI_STATUS_ITEM, Cardinal(@StatusParms));
  Result := StatusParms.dwReturn;
  { Volume: 0 - 1000 }
  end;

begin
  SoundAvailable:=WaveOutGetNumDevs>0;
  end.
