(* Erweiterung der Unit Masks durch Filtern mit erweiterten Ausdrcken
   ===================================================================
   
    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.
 
   Vers. 1 - Dez. 2004
   Vers. 2 - Aug. 2004 : multiple standard filter (f1|f2|..) added
   *)

unit TextFilter;

interface

uses SysUtils, Masks, RegExpr;

const
  defMatchAll = '*';
  defRegExpAll = '.';
  FilterSeparator = '|';

type
  TFilterMode = (fmOff,fmInclude,fmExclude);

{ ---------------------------------------------------------------- }
  TTextFilter = class (TRegExpr)
  private
    FRegExp    : boolean;
    FMode      : TFilterMode;
    FMask      : string;
    function GetFilter : string;
    procedure SetFilter (AFilter : string);
    function GetDefault : string;
    function ReadNextFilter (var AFilter : string) : string;
  public
    constructor Create;
    property Filter : string read GetFilter write SetFilter;
    property Mode : TFilterMode read FMode write FMode;
    property Regular : boolean read FRegExp write FRegExp;
    property MatchAll : string read GetDefault;
    function MatchesFilter(AText : string) : boolean;
    end;

{ ---------------------------------------------------------------- }
  EFilterError = class(Exception)
  public
    constructor Create (ErrString : string);
    end;

{ ---------------------------------------------------------------- }
implementation

{ ------------------------------------------------------------------- }
(* generate error message *)
constructor EFilterError.Create (ErrString : string);
begin
  inherited Create (ErrString);
  end;

{ ---------------------------------------------------------------- }
constructor TTextFilter.Create;
begin
  inherited Create;
  FRegExp:=false; FMode:=fmOff; FMask:=MatchAll;
  end;

function TTextFilter.GetDefault: string;
begin
  if FRegExp then Result:=defRegExpAll else Result:=defMatchAll;
  end;

function TTextFilter.ReadNextFilter (var AFilter : string) : string;
var
  i : integer;
begin
  if length(AFilter)>0 then begin
    i:=pos (FilterSeparator,AFilter);
    if i=0 then i:=succ(length(AFilter));
    Result:=copy(AFilter,1,pred(i));
    delete(AFilter,1,i);
    end
  else Result:='';
  end;

function TTextFilter.GetFilter : string;
begin
  if FRegExp then Result:=Expression else Result:=FMask;
  end;

procedure TTextFilter.SetFilter (AFilter : string);
begin
  if FRegExp then begin
    if length(AFilter)=0 then Expression:=defRegExpAll
    else begin
      Expression:=AFilter;
      try
        Exec('xyz');
      except
        Expression:=defRegExpAll;
        raise EFilterError.Create ('Illegal regular expression!');
        end;
      end;
    end
  else begin
    if length(AFilter)=0 then FMask:=defMatchAll
    else begin
      FMask:=AFilter;
      try
        repeat
          MatchesMask('xyz',ReadNextFilter(AFilter));
          until length(AFilter)=0;
      except
        FMask:=defMatchAll;
        raise EFilterError.Create ('Illegal mask!');
        end;
      end;
    end;
  end;

function TTextFilter.MatchesFilter(AText : string) : boolean;
var
  s : string;
begin
  if FMode=fmOff then Result:=true
  else begin
    if FRegExp then Result:=Exec(AText)
    else begin
      s:=FMask;
      repeat
        Result:=MatchesMask(AText,ReadNextFilter(s));
        until Result or (length(s)=0);
      end;
    if FMode=fmExclude then Result:=not Result;
    end;
  end;

end.
